"""Provide some function for creating and modifying channels (dataframes)"""
from __future__ import annotations

from glob import glob
from os import path
from typing import Any, List

import pandas
from pandas import DataFrame
from pipda import register_verb


# ----------------------------------------------------------------
# Creators
class Channel(DataFrame):
    """A DataFrame wrapper with creators"""

    @classmethod
    def create(cls, value: DataFrame | List[Any]) -> DataFrame:
        """Create a channel from a list.

        The second dimension is identified by tuple. if all elements are tuple,
        then a channel is created directly. Otherwise, elements are converted
        to tuples first and channels are created then.

        Examples:
            >>> Channel.create([1, 2, 3]) # 3 rows, 1 column
            >>> Channel.create([(1,2,3)]) # 1 row, 3 columns

        Args:
            value: The value to create a channel

        Returns:
            A channel (dataframe)
        """
        if isinstance(value, DataFrame):
            return value
        if all(isinstance(elem, tuple) for elem in value):
            return cls(value)
        return cls((val,) for val in value)

    @classmethod
    def from_glob(
        cls,
        pattern: str,
        ftype: str = "any",
        sortby: str = "name",
        reverse: bool = False,
    ) -> DataFrame:
        """Create a channel with a glob pattern

        Args:
            ftype: The file type, one of any, link, dir and file
            sortby: How the files should be sorted. One of name, mtime and size
            reverse: Whether sort them in a reversed way.

        Returns:
            The channel
        """
        sort_key = (
            str
            if sortby == "name"
            else path.getmtime
            if sortby == "mtime"
            else path.getsize
            if sortby == "size"
            else None
        )
        file_filter = (
            path.islink
            if ftype == "link"
            else path.isdir
            if ftype == "dir"
            else path.isfile
            if ftype == "file"
            else None
        )
        files = (
            file
            for file in glob(str(pattern))
            if not file_filter or file_filter(file)
        )
        return cls.create(
            sorted(files, key=sort_key, reverse=reverse),  # type: ignore
        )

    @classmethod
    def from_pairs(
        cls,
        pattern: str,
        ftype: str = "any",
        sortby: str = "name",
        reverse: bool = False,
    ) -> DataFrame:
        """Create a width=2 channel with a glob pattern

        Args:
            ftype: The file type, one of any, link, dir and file
            sortby: How the files should be sorted. One of name, mtime and size
            reverse: Whether sort them in a reversed way.

        Returns:
            The channel
        """
        mates = cls.from_glob(pattern, ftype, sortby, reverse)
        return pandas.concat(
            (
                mates.iloc[::2].reset_index(drop=True),
                mates.iloc[1::2].reset_index(drop=True),
            ),
            axis=1,
        )

    @classmethod
    def from_csv(cls, *args, **kwargs):
        """Create a channel from a csv file

        Uses pandas.read_csv() to create a channel

        Args:
            *args: and
            **kwargs: Arguments passing to pandas.read_csv()
        """
        return pandas.read_csv(*args, **kwargs)

    @classmethod
    def from_excel(cls, *args, **kwargs):
        """Create a channel from an excel file.

        Uses pandas.read_excel() to create a channel

        Args:
            *args: and
            **kwargs: Arguments passing to pandas.read_excel()
        """
        return pandas.read_excel(*args, **kwargs)

    @classmethod
    def from_table(cls, *args, **kwargs):
        """Create a channel from a table file.

        Uses pandas.read_table() to create a channel

        Args:
            *args: and
            **kwargs: Arguments passing to pandas.read_table()
        """
        return pandas.read_table(*args, **kwargs)


# ----------------------------------------------------------------
# Verbs
@register_verb(DataFrame)
def expand_dir(
    data: DataFrame,
    col: str | int = 0,
    pattern: str = "*",
    ftype: str = "any",
    sortby: str = "name",
    reverse: bool = False,
) -> DataFrame:
    """Expand a Channel according to the files in <col>,
    other cols will keep the same.

    This is only applicable to a 1-row channel.

    Examples:
        >>> ch = channel.create([('./', 1)])
        >>> ch >> expand()
        >>> [['./a', 1], ['./b', 1], ['./c', 1]]

    Args:
        col: the index or name of the column used to expand
        pattern: use a pattern to filter the files/dirs, default: `*`
        ftype: the type of the files/dirs to include
            - 'dir', 'file', 'link' or 'any' (default)
        sortby:  how the list is sorted
            - 'name' (default), 'mtime', 'size'
        reverse: reverse sort.

    Returns:
        The expanded channel
    """
    assert data.shape[0] == 1, "Can only expand a single row DataFrame."
    col_loc = col if isinstance(col, int) else data.columns.get_loc(col)
    full_pattern = f"{data.iloc[0, col_loc]}/{pattern}"
    expanded = Channel.from_glob(full_pattern, ftype, sortby, reverse)
    ret = pandas.concat([data] * expanded.size, axis=0)
    ret[data.columns[col_loc]] = expanded.values
    return ret.reset_index(drop=True)


@register_verb(DataFrame)
def collapse_files(data: DataFrame, col: str | int = 0) -> DataFrame:
    """Collapse a Channel according to the files in <col>,
    other cols will use the values in row 0.

    Note that other values in other rows will be discarded.

    Examples:
        >>> ch = channel.create([['./a', 1], ['./b', 1], ['./c', 1]])
        >>> ch >> collapse()
        >>> [['.', 1]]

    Args:
        data: The original channel
        col: the index or name of the column used to collapse on

    Returns:
        The collapsed channel
    """
    assert data.shape[0] > 0, "Cannot collapse on an empty DataFrame."
    col_loc = col if isinstance(col, int) else data.columns.get_loc(col)
    paths = list(data.iloc[:, col_loc])
    compx = path.dirname(path.commonprefix(paths))
    ret = data.iloc[[0], :].copy()
    ret.iloc[0, col_loc] = compx
    return ret



"""Provide builting schedulers"""
from __future__ import annotations

from typing import Type

from xqute import Scheduler
from xqute.schedulers.local_scheduler import LocalJob as XquteLocalJob
from xqute.schedulers.local_scheduler import (
    LocalScheduler as XquteLocalScheduler,
)
from xqute.schedulers.sge_scheduler import SgeJob as XquteSgeJob
from xqute.schedulers.sge_scheduler import SgeScheduler as XquteSgeScheduler

from .defaults import SCHEDULER_ENTRY_GROUP
from .exceptions import NoSuchSchedulerError, WrongSchedulerTypeError
from .job import Job
from .utils import is_subclass, load_entrypoints


class LocalJob(Job):
    """Job class for local scheduler"""

    wrap_cmd = XquteLocalJob.wrap_cmd


class LocalScheduler(XquteLocalScheduler):
    """Local scheduler"""

    job_class = LocalJob


class SgeJob(Job):
    """Job class for sge scheduler"""

    wrap_cmd = XquteSgeJob.wrap_cmd


class SgeScheduler(XquteSgeScheduler):
    """Sge scheduler"""

    job_class = SgeJob


def get_scheduler(scheduler: str | Type[Scheduler]) -> Type[Scheduler]:
    """Get the scheduler by name of the scheduler class itself

    Args:
        scheduler: The scheduler class or name

    Returns:
        The scheduler class
    """
    if is_subclass(scheduler, Scheduler):
        return scheduler  # type: ignore

    if scheduler == "local":
        return LocalScheduler
    if scheduler == "sge":
        return SgeScheduler

    for name, obj in load_entrypoints(
        SCHEDULER_ENTRY_GROUP
    ):  # pragma: no cover
        if name == scheduler:
            if not is_subclass(obj, Scheduler):
                raise WrongSchedulerTypeError(
                    "Scheduler should be a subclass of "
                    "pipen.scheduler.Scheduler."
                )
            return obj

    raise NoSuchSchedulerError(str(scheduler))



"""Process group that contains a set of processes.

It can be easily used to create a pipeline that runs independently or
integrated into a larger pipeline.

Runs directly:
>>> proc_group = ProcGroup(<options>)
>>> proc_group.as_pipen(<pipeline options>).set_data(<data>).run()

Integrated into a larger pipeline
>>> proc_group = ProcGroup(<options>)
>>> # proc could be a process within the larger pipeline
>>> proc.requires = prog_group.<proc>

To add a process to the proc group, use the `add_proc` method:
>>> class MyProcGroup(ProcGroup):
>>>     ...
>>>
>>> proc_group = MyProcGroup(...)
>>> @proc_group.add_proc
>>> class MyProc(Proc):
>>>     ...

Or add a process at runtime:
>>> class MyProcGroup(ProcGroup):
>>>     ...
>>>
>>>     @ProcGroup.add_proc
>>>     def my_proc(self):
>>>         class MyProc(Proc):
>>>             # You may use self.options here
>>>             ...
>>>         return MyProc
>>> proc_group = MyProcGroup(...)
"""
from __future__ import annotations

from os import PathLike
from functools import wraps
from typing import Callable, Type, List
from abc import ABC, ABCMeta
from diot import Diot

from .utils import cached_property
from .pipen import Pipen
from .proc import Proc


class ProcGropuMeta(ABCMeta):
    """Meta class for ProcGroup"""

    _INST = None

    def __call__(cls, *args, **kwds):
        """Make sure Proc subclasses are singletons

        Args:
            *args: and
            **kwds: Arguments for the constructor

        Returns:
            The Proc instance
        """
        if cls._INST is None:
            cls._INST = super().__call__(*args, **kwds)

        return cls._INST


class ProcGroup(ABC, metaclass=ProcGropuMeta):
    """A group of processes that can be run independently or
    integrated into a larger pipeline.
    """

    name: str | None = None
    DEFAULTS = Diot()
    PRESERVED = {
        "opts",
        "name",
        "add_proc",
        "as_pipen",
        "procs",
        "starts",
        "DEFAULTS",
        "PRESERVED",
        "_INST",
    }

    def __init__(self, **opts) -> None:
        self.opts = Diot(self.__class__.DEFAULTS or {}) | (opts or {})
        self.name = self.__class__.name or self.__class__.__name__
        self.starts: List[Type[Proc]] = []
        self.procs = Diot()

        self._load_runtime_procs()

    def _load_runtime_procs(self):
        """Load all processes that are added at runtime"""
        # Load all processes if they are decorated by ProcGroup.add_proc
        for name, attr in self.__class__.__dict__.items():
            if isinstance(attr, cached_property):
                getattr(self, name)

    def add_proc(
        self_or_method: ProcGroup | Callable[[ProcGroup], Type[Proc]],
        proc: Type[Proc] | None = None,
    ) -> Type[Proc] | cached_property:
        """Add a process to the proc group

        It works either as a decorator to the process directly or as a
        decorator to a method that returns the process.

        Args:
            self_or_method: The proc group instance or a method that
                returns the process
            proc: The process class if `self_or_method` is the proc group

        Returns:
            The process class if `self_or_method` is the proc group, or
            a cached property that returns the process class
        """
        if isinstance(self_or_method, ProcGroup):
            # Called as self.add_proc or pg.add_proc
            if proc is None:
                return self_or_method.add_proc  # type: ignore

            if proc.name in self_or_method.__class__.PRESERVED:
                raise ValueError(
                    f"Process name `{proc.name}` is reserved for ProcGroup"
                )

            setattr(self_or_method, proc.name, proc)
            proc.__procgroup__ = self_or_method
            if not proc.requires:
                self_or_method.starts.append(proc)
            self_or_method.procs[proc.name] = proc
            return proc

        @wraps(self_or_method)
        def wrapper(self):
            proc = self_or_method(self)

            if proc is None:
                return None

            if (
                not isinstance(proc, Proc)
                and (not isinstance(proc, type) or not issubclass(proc, Proc))
            ):
                raise ValueError(
                    f"`{proc}` is not a process "
                    "(either a subclass or an instance of Proc)"
                )

            proc.__procgroup__ = self
            if not proc.requires:
                self.starts.append(proc)
            self.procs[proc.name] = proc
            return proc

        return cached_property(wrapper)

    def as_pipen(
        self,
        name: str | None = None,
        desc: str | None = None,
        outdir: str | PathLike | None = None,
        **kwargs,
    ) -> Pipen:
        """Convert the pipeline to a Pipen instance

        Args:
            name: The name of the pipeline
            desc: The description of the pipeline
            outdir: The output directory of the pipeline
            **kwargs: The keyword arguments to pass to Pipen

        Returns:
            The Pipen instance
        """
        name = name or self.__class__.__name__
        if self.__doc__:
            desc = desc or self.__doc__.lstrip().splitlines()[0]

        pipe = Pipen(name=name, desc=desc, outdir=outdir, **kwargs)
        pipe.set_start(self.starts)
        return pipe
